#include "main.h"

#define BUFFER_COUNT 30

typedef struct {
	int fd; // fd from NVBufSurface
	VkDeviceMemory vkmem; // memory from vulkan
} CompleteBufferContext;

typedef struct {
	RenderCtx* ctx;
	int index;
} ThreadArgs;

CompleteBufferContext* init_buffers(RenderCtx* ctx, int buffer_count, uint32_t width, uint32_t height, int index)
{
	CompleteBufferContext* buffers = (CompleteBufferContext*) calloc(buffer_count, sizeof(CompleteBufferContext));
	if (buffers)
	{
		for (uint32_t i = 0; i < buffer_count; i++)
		{
			CK_BOOL(nvbufsurface_init(&buffers[i].fd, width, height, index));
		}
		for (uint32_t i = 0; i < buffer_count; i++)
		{
			CK_BOOL(render_import_dma_fd(ctx, buffers[i].fd, &(buffers[i].vkmem), index));
		}
	}

	return buffers;
error:
	return NULL;
}

void fini_buffers(RenderCtx* ctx, CompleteBufferContext* buffers, int buffer_count, int index)
{
	for (uint32_t i = 0; i < buffer_count; i++)
	{
		// nvbufsurface_fini(buffers[i].fd, index);
		render_fini_device_memory(ctx, buffers[i].vkmem); // Vulkan gains ownership of the buffers, so we do not free with nvbufsurface.
	}
}

void do_the_thing(RenderCtx* ctx, int index)
{
	static bool start = true;
	if (start)
	{
		start = false;
		usleep(10000); // sleeping here on the first thread call to make sure both threads run through the same code at the same time.
	}
	CompleteBufferContext* buffers = init_buffers(ctx, BUFFER_COUNT, 3840, 2160, index);
	if (buffers)
	{
		sleep(1);
		fini_buffers(ctx, buffers, BUFFER_COUNT, index);
		free(buffers);
	}
}

void* myfunc(void* args)
{
	do_the_thing(((ThreadArgs*)args)->ctx, ((ThreadArgs*)args)->index);
}

int main(void)
{
	RenderCtx* ctx = render_init(true);
	if (ctx)
	{
		ThreadArgs args1 = {0};
		ThreadArgs args2 = {0};
		pthread_t thread1;
		pthread_t thread2;
		args1.ctx = ctx;
		args2.ctx = ctx;
		args1.index = 0;
		args2.index = 1;
		pthread_create(&thread1, NULL, myfunc, &args1);
		pthread_create(&thread2, NULL, myfunc, &args2);
		pthread_join(thread1, NULL);
		pthread_join(thread2, NULL);
		render_fini(ctx);
	}

	return 0;
error:
	return -1;
}
