#include "render.ph"

#define VK_ENUMERATE_NO_CK(array, type, function)\
{\
	function;\
	array = (type*) calloc(count, sizeof(type));\
	CK_PTR(array);\
	function;\
}

#define VK_ENUMERATE(array, type, function)\
{\
	result = function;\
	CK_VK(result);\
	array = (type*) calloc(count, sizeof(type));\
	CK_PTR(array);\
	result = function;\
	CK_VK(result);\
}

bool render_enumerate_layers(RenderCtx* ctx)
{
	VkResult result;
	uint32_t count = 0;

	CK_PTR(ctx);

	VK_ENUMERATE(ctx->layer_properties, VkLayerProperties, vkEnumerateInstanceLayerProperties(&count, ctx->layer_properties));

	// for (uint32_t i = 0; i < count; i ++)
	// {
	// 	LOG_I("                : %s", ctx->layer_properties[i].layerName);
	// }
	ctx->layer_properties_count = count;

	return true;
error:
	return false;
}

bool render_enumerate_extensions(RenderCtx* ctx, const char* layer)
{
	VkResult result;
	uint32_t count = 0;

	CK_PTR(ctx);

	free(ctx->extension_properties);
	ctx->extension_properties = NULL;

	VK_ENUMERATE(ctx->extension_properties, VkExtensionProperties, vkEnumerateInstanceExtensionProperties(layer, &count, ctx->extension_properties));

	ctx->extension_properties_count = count;

	// for (uint32_t i = 0; i < count; i ++)
	// {
	// 	LOG_I("                : %s", ctx->extension_properties[i].extensionName);
	// }

	return true;
error:
	return false;
}

bool render_enumerate_physical_devices(RenderCtx* ctx)
{
	VkResult result;
	uint32_t count = 0;

	CK_PTR(ctx);

	VK_ENUMERATE(ctx->physical_devices, VkPhysicalDevice, vkEnumeratePhysicalDevices(ctx->instance, &count, ctx->physical_devices));
	for (uint32_t i = 0; i < count; i ++)
	{
		VkPhysicalDeviceProperties props = {0};
		vkGetPhysicalDeviceProperties(ctx->physical_devices[i], &props);

		if (props.deviceType == VK_PHYSICAL_DEVICE_TYPE_INTEGRATED_GPU)
		{
			ctx->active_physical_device = i;
			LOG_I("Selected GPU : %s", props.deviceName);
			break;
		}
		else
		{
			LOG_I("                : %s", props.deviceName);
		}
	}
	ctx->physical_device_count = count;

	return true;
error:
	return false;
}

bool render_enumerate_queue_family_properties(RenderCtx* ctx, VkPhysicalDevice device)
{
	uint32_t count = 0;

	CK_PTR(ctx);

	VK_ENUMERATE_NO_CK(ctx->queue_family_properties, VkQueueFamilyProperties, vkGetPhysicalDeviceQueueFamilyProperties(device, &count, ctx->queue_family_properties));
	// LOG_I("Physical Devices:");
	// for (uint32_t i = 0; i < count; i ++)
	// {
	// 	if (ctx->queue_family_properties[i].queueFlags & VK_QUEUE_GRAPHICS_BIT)
	// 	{
	// 		LOG_I("Graphics Queue %u", i);
	// 	}
	// }
	ctx->queue_family_properties_count = count;

	return true;
error:
	return false;
}

bool render_enumerate_device_extensions(RenderCtx* ctx, VkPhysicalDevice device)
{
	VkResult result;
	uint32_t count = 0;

	CK_PTR(ctx);

	VK_ENUMERATE(ctx->device_extension_properties, VkExtensionProperties, vkEnumerateDeviceExtensionProperties(device, NULL, &count, ctx->device_extension_properties));
	for (uint32_t i = 0; i < count; i++)
	{

	}
	ctx->device_extension_properties_count = count;

	// for (uint32_t i = 0; i < count; i ++)
	// {
	// 	LOG_I("                : %s", ctx->device_extension_properties[i].extensionName);
	// }

	return true;
error:
	return false;
}

bool render_enumerate_surface_formats(RenderCtx* ctx, VkPhysicalDevice device)
{
	VkResult result;
	uint32_t count = 0;

	VK_ENUMERATE(ctx->surface_formats, VkSurfaceFormatKHR, vkGetPhysicalDeviceSurfaceFormatsKHR(device, ctx->surface, &count, ctx->surface_formats));
	ctx->surface_format_count = count;

	return true;
error:
	return false;
}

bool render_enumerate_present_modes(RenderCtx* ctx, VkPhysicalDevice device)
{
	VkResult result;
	uint32_t count = 0;

	VK_ENUMERATE(ctx->present_modes, VkPresentModeKHR, vkGetPhysicalDeviceSurfacePresentModesKHR(device, ctx->surface, &count, ctx->present_modes));
	ctx->present_mode_count = count;

	return true;
error:
	return false;
}

void render_push_layer(RenderCtx* ctx, char* layer_name)
{
	if (render_find_layer(ctx, layer_name))
	{
		if (ctx->layers)
		{
			// inefficient, but clean. this only is needed for init, so no big deal.
			char** temp = (char**) calloc(ctx->layers_count + 1, sizeof(char*));
			if (!temp)
			{
				LOG_E("Failed to allocate memory to store extension: %s", layer_name);
			}
			else
			{
				memcpy(temp, ctx->layers, sizeof(char*) * ctx->layers_count);
				free(ctx->layers);
				ctx->layers = temp;
				ctx->layers[ctx->layers_count] = layer_name;
				ctx->layers_count++;
			}
		}
		else
		{
			ctx->layers = (char**) calloc(1, sizeof(char*));
			ctx->layers[ctx->layers_count] = layer_name;
			ctx->layers_count++;
		}
	}
	else
	{
		LOG_W("Failed to find layer: %s", layer_name);
	}
}

void render_push_extension(RenderCtx* ctx, char* extension_name)
{
	if (render_find_extension(ctx, extension_name))
	{
		if (ctx->extensions)
		{
			// inefficient, but clean. this only is needed for init, so no big deal.
			char** temp = (char**) calloc(ctx->extensions_count + 1, sizeof(char*));
			if (!temp)
			{
				LOG_E("Failed to allocate memory to store extension: %s", extension_name);
			}
			else
			{
				memcpy(temp, ctx->extensions, sizeof(char*) * ctx->extensions_count);
				free(ctx->extensions);
				ctx->extensions = temp;
				ctx->extensions[ctx->extensions_count] = extension_name;
				ctx->extensions_count++;
			}
		}
		else
		{
			ctx->extensions = (char**) calloc(1, sizeof(char*));
			ctx->extensions[ctx->extensions_count] = extension_name;
			ctx->extensions_count++;
		}
	}
	else
	{
		LOG_W("Failed to find extension: %s", extension_name);
	}
}

void render_push_device_extension(RenderCtx* ctx, char* extension_name)
{
	if (render_find_device_extension(ctx, extension_name))
	{
		if (ctx->device_extensions)
		{
			// inefficient, but clean. this only is needed for init, so no big deal.
			char** temp = (char**) calloc(ctx->device_extensions_count + 1, sizeof(char*));
			if (!temp)
			{
				LOG_E("Failed to allocate memory to store extension: %s", extension_name);
			}
			else
			{
				memcpy(temp, ctx->device_extensions, sizeof(char*) * ctx->device_extensions_count);
				free(ctx->device_extensions);
				ctx->device_extensions = temp;
				ctx->device_extensions[ctx->device_extensions_count] = extension_name;
				ctx->device_extensions_count++;
			}
		}
		else
		{
			ctx->device_extensions = (char**) calloc(1, sizeof(char*));
			ctx->device_extensions[ctx->device_extensions_count] = extension_name;
			ctx->device_extensions_count++;
		}
	}
	else
	{
		LOG_W("Find to find device extension: %s", extension_name);
	}
}

bool render_find_layer(RenderCtx* ctx, char* layer_name)
{
	bool found = false;

	CK_PTR(ctx);
	CK_PTR(layer_name);

	for (uint32_t i = 0; i < ctx->layer_properties_count; i++)
	{
		if (strcmp(ctx->layer_properties[i].layerName, layer_name) == 0)
		{
			found = true;
			return true;
		}
	}

error:
	return false;
}

bool render_find_extension(RenderCtx* ctx, char* extension_name)
{
	bool found = false;

	CK_PTR(ctx);
	CK_PTR(extension_name);

	for (uint32_t i = 0; i < ctx->extension_properties_count; i++)
	{
		if (strcmp(ctx->extension_properties[i].extensionName, extension_name) == 0)
		{
			found = true;
			break;
		}
	}

error:
	return found;
}

bool render_find_queue_families(RenderCtx* ctx)
{
	VkResult result;

	CK_PTR(ctx);

	VkBool32 presentable = false;
	for (uint32_t i = 0; i < ctx->queue_family_properties_count; i++)
	{
		if (ctx->queue_family_properties[i].queueFlags & VK_QUEUE_GRAPHICS_BIT)
		{
			ctx->queue_indices.graphics = i;
		}
	}

	return true;
error:
	return false;
}

bool render_find_device_extension(RenderCtx* ctx, char* extension_name)
{
	bool found = false;

	CK_PTR(ctx);

	for (uint32_t i = 0; i < ctx->device_extension_properties_count; i++)
	{
		if (strcmp(ctx->device_extension_properties[i].extensionName, extension_name) == 0)
		{
			found = true;
			break;
		}
	}

error:
	return found;
}
