#pragma once

//render.ph (private header)

#define FABRYK_RENDER_PH

#include <stdbool.h>

#include <stdlib.h>
#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include "../util/logger.h"

#define CK_VK(result)\
	if (result != VK_SUCCESS)\
	{\
		render_log_vk_result(result);\
		goto error;\
	}

typedef struct RenderCtx RenderCtx;
typedef struct RenderQueueFamilies RenderQueueFamilies;
typedef struct RenderExtensionProcs RenderExtensionProcs;

#include "render.h"

struct RenderQueueFamilies {
	int32_t graphics;
	int32_t present;
};

struct RenderExtensionProcs {
	PFN_vkCreateDebugUtilsMessengerEXT vkCreateDebugUtilsMessengerEXT;
	PFN_vkDestroyDebugUtilsMessengerEXT vkDestroyDebugUtilsMessengerEXT;
};

struct RenderCtx {
	VkInstance instance;
	VkDevice device;
	VkSurfaceKHR surface;
	VkQueue graphics_queue;
	VkQueue present_queue;

	RenderQueueFamilies queue_indices;

	char** layers;
	uint32_t layers_count;
	VkLayerProperties* layer_properties;
	uint32_t layer_properties_count;

	char** extensions;
	uint32_t extensions_count;
	VkExtensionProperties* extension_properties;
	uint32_t extension_properties_count;

	char** queue_families;
	uint32_t queue_families_count;
	VkQueueFamilyProperties* queue_family_properties;
	uint32_t queue_family_properties_count;

	char** device_extensions;
	uint32_t device_extensions_count;
	VkExtensionProperties* device_extension_properties;
	uint32_t device_extension_properties_count;

	VkSurfaceFormatKHR* surface_formats;
	uint32_t surface_format_count;
	uint32_t active_surface_format;

	VkPresentModeKHR* present_modes;
	uint32_t present_mode_count;
	uint32_t active_present_mode;

	RenderExtensionProcs procs;

	uint32_t physical_device_count;
	uint32_t active_physical_device;
	VkPhysicalDevice* physical_devices;

	bool enable_validation_layers;
	VkDebugUtilsMessengerEXT debug_messenger;
	VkAllocationCallbacks* allocators;
};

bool render_init_layers(RenderCtx* ctx);
bool render_init_extensions(RenderCtx* ctx);
bool render_init_instance(RenderCtx* ctx);
bool render_init_surface(RenderCtx* ctx);
bool render_init_logical_device(RenderCtx* ctx);
bool render_init_device_extensions(RenderCtx* ctx);
bool render_init_swapchain(RenderCtx* ctx);
bool render_init_debug_messenger(RenderCtx* ctx);

void render_fini_debug_messenger(RenderCtx* ctx);

bool render_enumerate_layers(RenderCtx* ctx);
bool render_enumerate_extensions(RenderCtx* ctx, const char* layer);
bool render_enumerate_physical_devices(RenderCtx* ctx);
bool render_enumerate_queue_family_properties(RenderCtx* ctx, VkPhysicalDevice device);
bool render_enumerate_device_extensions(RenderCtx* ctx, VkPhysicalDevice device);
bool render_enumerate_surface_formats(RenderCtx* ctx, VkPhysicalDevice device);
bool render_enumerate_present_modes(RenderCtx* ctx, VkPhysicalDevice device);

void render_push_layer(RenderCtx* ctx, char* layer_name);
void render_push_extension(RenderCtx* ctx, char* extension_name);
void render_push_device_extension(RenderCtx* ctx, char* extension_name);

bool render_find_layer(RenderCtx* ctx, char* layer_name);
bool render_find_extension(RenderCtx* ctx, char* extension_name);
bool render_find_queue_families(RenderCtx* ctx);
bool render_find_device_extension(RenderCtx* ctx, char* extension_name);

bool render_init_procs(RenderCtx* ctx);

VkDebugUtilsMessengerCreateInfoEXT render_get_debug_create_info();
static VKAPI_ATTR VkBool32 VKAPI_CALL render_debug_callback(VkDebugUtilsMessageSeverityFlagBitsEXT message_severity, VkDebugUtilsMessageTypeFlagsEXT message_type, const VkDebugUtilsMessengerCallbackDataEXT* callback_data, void* user_data);
const char* render_log_validation_layer_with_type(VkDebugUtilsMessageTypeFlagsEXT message_type);

void render_log_vk_result(VkResult result);

#include "app_info.h"
