#include "render.ph"

bool render_init_debug_messenger(RenderCtx* ctx)
{
	CK_PTR(ctx);
	CK_PTR(ctx->procs.vkCreateDebugUtilsMessengerEXT);

	VkResult result;
	VkDebugUtilsMessengerCreateInfoEXT create_info = render_get_debug_create_info();

	create_info.sType = VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CREATE_INFO_EXT;
	create_info.messageSeverity =
		VK_DEBUG_UTILS_MESSAGE_SEVERITY_VERBOSE_BIT_EXT |
		VK_DEBUG_UTILS_MESSAGE_SEVERITY_INFO_BIT_EXT |
		VK_DEBUG_UTILS_MESSAGE_SEVERITY_WARNING_BIT_EXT |
		VK_DEBUG_UTILS_MESSAGE_SEVERITY_ERROR_BIT_EXT;
	create_info.messageType =
		VK_DEBUG_UTILS_MESSAGE_TYPE_GENERAL_BIT_EXT |
		VK_DEBUG_UTILS_MESSAGE_TYPE_VALIDATION_BIT_EXT |
		VK_DEBUG_UTILS_MESSAGE_TYPE_PERFORMANCE_BIT_EXT;
	create_info.pfnUserCallback = render_debug_callback;
	create_info.pUserData = NULL;

	result = ctx->procs.vkCreateDebugUtilsMessengerEXT(ctx->instance, &create_info, ctx->allocators, &ctx->debug_messenger);
	CK_VK(result);

	return true;
error:
	return false;
}

void render_fini_debug_messenger(RenderCtx* ctx)
{
	CK_PTR(ctx);
	CK_PTR(ctx->procs.vkDestroyDebugUtilsMessengerEXT);

	ctx->procs.vkDestroyDebugUtilsMessengerEXT(ctx->instance, ctx->debug_messenger, ctx->allocators);

	return;
error:
	LOG_E("Failed to fini debug messenger");
	return;
}

VkDebugUtilsMessengerCreateInfoEXT render_get_debug_create_info()
{
	VkDebugUtilsMessengerCreateInfoEXT create_info = {0};

	create_info.sType = VK_STRUCTURE_TYPE_DEBUG_UTILS_MESSENGER_CREATE_INFO_EXT;
	create_info.messageSeverity =
		VK_DEBUG_UTILS_MESSAGE_SEVERITY_VERBOSE_BIT_EXT |
		VK_DEBUG_UTILS_MESSAGE_SEVERITY_INFO_BIT_EXT |
		VK_DEBUG_UTILS_MESSAGE_SEVERITY_WARNING_BIT_EXT |
		VK_DEBUG_UTILS_MESSAGE_SEVERITY_ERROR_BIT_EXT;
	create_info.messageType =
		VK_DEBUG_UTILS_MESSAGE_TYPE_GENERAL_BIT_EXT |
		VK_DEBUG_UTILS_MESSAGE_TYPE_VALIDATION_BIT_EXT |
		VK_DEBUG_UTILS_MESSAGE_TYPE_PERFORMANCE_BIT_EXT;
	create_info.pfnUserCallback = render_debug_callback;
	create_info.pUserData = NULL;

	return create_info;
}

static VKAPI_ATTR VkBool32 VKAPI_CALL render_debug_callback(
	VkDebugUtilsMessageSeverityFlagBitsEXT message_severity,
	VkDebugUtilsMessageTypeFlagsEXT message_type,
	const VkDebugUtilsMessengerCallbackDataEXT* callback_data,
	void* user_data)
{
	user_data = user_data;

	const char* type = render_log_validation_layer_with_type(message_type);
	switch (message_severity)
	{
		case VK_DEBUG_UTILS_MESSAGE_SEVERITY_VERBOSE_BIT_EXT:
			LOG_I("Validation Layers (%s): %s", type, callback_data->pMessage);
			break;
		case VK_DEBUG_UTILS_MESSAGE_SEVERITY_INFO_BIT_EXT:
			LOG_I("Validation Layers (%s): %s", type, callback_data->pMessage);
			break;
		case VK_DEBUG_UTILS_MESSAGE_SEVERITY_WARNING_BIT_EXT:
			LOG_W("Validation Layers (%s): %s", type, callback_data->pMessage);
			break;
		case VK_DEBUG_UTILS_MESSAGE_SEVERITY_ERROR_BIT_EXT:
			LOG_E("Validation Layers (%s): %s", type, callback_data->pMessage);
			break;
		default:
			LOG_E("Validation Layers with unknown severity (%s): %s", type, callback_data->pMessage);
			break;
	}

	return VK_FALSE;
}

const char* render_log_validation_layer_with_type(VkDebugUtilsMessageTypeFlagsEXT message_type)
{
	const char* type;
	switch (message_type)
	{
		case VK_DEBUG_UTILS_MESSAGE_TYPE_GENERAL_BIT_EXT:
			type = "General";
			break;
		case VK_DEBUG_UTILS_MESSAGE_TYPE_VALIDATION_BIT_EXT:
			type = "Validation";
			break;
		case VK_DEBUG_UTILS_MESSAGE_TYPE_PERFORMANCE_BIT_EXT:
			type = "Performance";
			break;
		default:
			type = "Unknown";
			break;
	}

	return type;
}
